@@ -9,6 +9,7 @@ from django.core.exceptions import PermissionDenied
 from django.core.urlresolvers import reverse
 from django.db import transaction
 from django.db.models import Q
+from django.http import HttpResponseRedirect
 from django.shortcuts import get_object_or_404
 from django.template import defaultfilters
 from django.utils.decorators import method_decorator
@@ -25,9 +26,9 @@ class CannedReplyAdmin(admin.ModelAdmin):
 
 class CommentReplyInline(admin.StackedInline):
     extra = 1
-    exclude = ['reply_comment', ]
+    exclude = ['reply_comments', ]
     model = models.CommentReply
-    fk_name = 'replied_to_comment'
+    fk_name = 'replied_to_comments'
 
     def formfield_for_foreignkey(self, db_field, request=None, **kwargs):
         """
@@ -43,8 +44,16 @@ class CommentReplyInline(admin.StackedInline):
 
 
 class CommentReplyAdmin(admin.ModelAdmin):
-    raw_id_fields = ("replied_to_comment", )
-    exclude = ("reply_comment", )
+    raw_id_fields = ("replied_to_comments", )
+    exclude = ("reply_comments", )
+
+    def response_add(self, request, obj, post_url_continue=None):
+        if 'admin_redirect' in request.session:
+            redirect = request.session['admin_redirect']
+            del request.session["admin_redirect"]
+            return HttpResponseRedirect(redirect)
+        else:
+            return super(CommentReplyAdmin, self).response_add(request, obj, post_url_continue)
 
 
 class CommentAdmin(DjangoCommentsAdmin):
@@ -55,84 +64,26 @@ class CommentAdmin(DjangoCommentsAdmin):
         '_user',
         'submit_date',
     )
-    inlines = [
-        CommentReplyInline,
-    ]
+    actions = ['mark_spam', 'mark_ham', 'add_moderator_reply']
 
     def queryset(self, request):
         """
         Exclude replies from listing since they are displayed inline as part of listing.
+
+        For proxy models with cls apptribute limit comments to those classified as cls.
         """
         qs = super(CommentAdmin, self).queryset(request)
-        return qs.filter(reply_comment_set__isnull=True)
+        qs = qs.filter(reply_comments_set__isnull=True)
+        cls = getattr(self, 'cls', None)
+        if cls:
+            qs = qs.filter(classifiedcomment__cls=self.cls)
+        return qs.select_related('user', 'content_type')
 
-    def get_changelist(self, request):
-        """
-        Used by AdminModeratorMixin.moderate_view to somewhat hackishly limit comments to only those
-        for the object under review, but only if an obj attribute is found on request
-        (which means the mixin is being applied and we are not on the standard changelist_view).
-        """
-
-        class ModeratorChangeList(ChangeList):
-            def get_query_set(self, request):
-                qs = super(ModeratorChangeList, self).get_query_set(request)
-                obj = getattr(request, 'obj', None)
-                if obj:
-                    ct = ContentType.objects.get_for_model(obj)
-                    qs = qs.filter(content_type=ct, object_pk=obj.pk)
-                return qs
-
-        return ModeratorChangeList
-
-    def content(self, obj):
-        content = obj.content_object
-        content_type = obj.content_type
-        url = reverse('admin:%s_%s_moderate' % (content_type.app_label,
-                                              content_type.model),
-                      args=(content.id,))
-
-        return '<a href="%s">%s</a>' % (url, content)
-    content.allow_tags = True
-
-    def moderator_reply(self, obj):
-        replies = obj.replied_to_comment_set.all()
-        if replies:
-            reply = replies[0]
-            change_icon = '<img src="%s" alt="change" />' % static('admin/img/icon_changelink.gif')
-            return '%s <a href="%s" target="_blank">%s</a>' % (
-                change_icon,
-                reverse('admin:moderator_commentreply_change', args=(reply.id, )),
-                reply.reply_comment.comment
-            )
-        else:
-            add_icon = '<img src="%s" alt="add" />' % static('admin/img/icon_addlink.gif')
-            return '%s <a href="%s?replied_to_comment=%s" class="add-another" id="moderator_reply_comment_%s" onclick="return showAddAnotherPopup(this);">Add reply</a>' % (
-                add_icon,
-                reverse('admin:moderator_commentreply_add'),
-                obj.pk,
-                obj.pk
-            )
-    moderator_reply.allow_tags = True
-
-    def comment_text(self, obj):
-        return '<a href="%s">%s</a>' % (
-            reverse('admin:comments_comment_change', args=(obj.id, )),
-            defaultfilters.linebreaks(obj.comment)
-        )
-    comment_text.short_description = 'Comment'
-    comment_text.allow_tags = True
-
-    def _user(self, obj):
-        url = reverse('admin:auth_user_change', args=(obj.user.id,))
-        return '<a href="%s">%s</a>' % (url, obj.user)
-    _user.allow_tags = True
-    _user.short_description = 'User'
-
-
-class CommentProxyAdmin(CommentAdmin):
-    def queryset(self, request):
-        qs = super(CommentProxyAdmin, self).queryset(request)
-        return qs.filter(classifiedcomment__cls=self.cls, reply_comment_set__isnull=True)
+    def add_moderator_reply(self, modeladmin, request, queryset):
+        selected = request.POST.getlist(admin.ACTION_CHECKBOX_NAME)
+        request.session["admin_redirect"] = request.get_full_path()
+        return HttpResponseRedirect(reverse('admin:moderator_commentreply_add') + '?replied_to_comments=%s' % ",".join(selected))
+    add_moderator_reply.short_description = "Add moderator reply"
 
     def mark_spam(self, modeladmin, request, queryset):
         for comment in queryset:
@@ -156,25 +107,119 @@ class CommentProxyAdmin(CommentAdmin):
             )
         return actions
 
+    def get_changelist(self, request):
+        class ModeratorChangeList(ChangeList):
+            def get_query_set(self, request):
+                """
+                Used by AdminModeratorMixin.moderate_view to somewhat hackishly limit comments to only those
+                for the object under review, but only if an obj attribute is found on request
+                (which means the mixin is being applied and we are not on the standard changelist_view).
+                """
+                qs = super(ModeratorChangeList, self).get_query_set(request)
+                obj = getattr(request, 'obj', None)
+                if obj:
+                    ct = ContentType.objects.get_for_model(obj)
+                    qs = qs.filter(content_type=ct, object_pk=obj.pk)
+                return qs
 
-class HamCommentAdmin(CommentProxyAdmin):
+            def get_results(self, request):
+                """
+                Create a content_type map to individual objects through their id's to avoid
+                additional per object queries for generic relation lookup (used in CommentAdmin.content
+                method).
+
+                Also create a comment_reply map to avoid additional reply lookups per comment
+                object (used in CommentAdmin.moderator_reply method)
+                """
+                super(ModeratorChangeList, self).get_results(request)
+                comment_ids = list(self.result_list.values_list('id', flat=True))
+                object_pks = list(self.result_list.values_list('object_pk', flat=True))
+
+                ct_map = {}
+                for obj in self.result_list:
+                    if obj.content_type not in ct_map:
+                        ct_map.setdefault(obj.content_type, {})
+                        for content_obj in obj.content_type.model_class()._default_manager.filter(pk__in=object_pks):
+                            ct_map[obj.content_type][content_obj.id] = content_obj
+                self.model_admin.ct_map = ct_map
+
+                comment_replies = {}
+                for comment_reply in models.CommentReply.objects.filter(replied_to_comments__in=comment_ids):
+                    for replied_to_comment in comment_reply.replied_to_comments.all():
+                        comment_replies[replied_to_comment.id] = {
+                            'id': comment_reply.id,
+                            'comment': comment_reply.comment_text,
+                        }
+                self.model_admin.comment_replies = comment_replies
+
+        return ModeratorChangeList
+
+    def content(self, obj, *args, **kwargs):
+        content_type = obj.content_type
+        content = self.ct_map[content_type][int(obj.object_pk)]
+        url = reverse('admin:%s_%s_moderate' % (content_type.app_label,
+                                              content_type.model),
+                      args=(content.id,))
+
+        return '<a href="%s">%s</a>' % (url, content)
+    content.allow_tags = True
+
+    def moderator_reply(self, obj):
+        if obj.id in self.comment_replies:
+            reply = self.comment_replies[obj.id]
+            change_icon = '<img src="%s" alt="change" />' % static('admin/img/icon_changelink.gif')
+            return '%s <a href="%s" target="_blank">%s</a>' % (
+                change_icon,
+                reverse('admin:moderator_commentreply_change', args=(reply['id'], )),
+                reply['comment']
+            )
+        else:
+            add_icon = '<img src="%s" alt="add" />' % static('admin/img/icon_addlink.gif')
+            return '%s <a href="%s?replied_to_comments=%s" class="add-another" id="moderator_reply_comments_%s" onclick="return showAddAnotherPopup(this);">Add reply</a>' % (
+                add_icon,
+                reverse('admin:moderator_commentreply_add'),
+                obj.pk,
+                obj.pk
+            )
+    moderator_reply.allow_tags = True
+
+    def comment_text(self, obj):
+        return '<a href="%s">%s</a>' % (
+            reverse(
+                'admin:%s_%s_change' % (obj._meta.app_label, obj._meta.module_name),
+                args=(obj.id, )
+            ),
+            defaultfilters.linebreaks(obj.comment)
+        )
+    comment_text.short_description = 'Comment'
+    comment_text.allow_tags = True
+
+    def _user(self, obj):
+        url = reverse('admin:auth_user_change', args=(obj.user.id,))
+        return '<a href="%s">%s</a>' % (url, obj.user)
+    _user.allow_tags = True
+    _user.short_description = 'User'
+
+
+class HamCommentAdmin(CommentAdmin):
     cls = 'ham'
-    actions = ['mark_spam', ]
+    actions = ['add_moderator_reply', 'mark_spam', ]
+    raw_id_fields = ('user', )
 
 
-class ReportedCommentAdmin(CommentProxyAdmin):
+class ReportedCommentAdmin(CommentAdmin):
     cls = 'reported'
-    actions = ['mark_ham', 'mark_spam', ]
+    actions = ['add_moderator_reply', 'mark_ham', 'mark_spam', ]
 
 
-class SpamCommentAdmin(CommentProxyAdmin):
+class SpamCommentAdmin(CommentAdmin):
     cls = 'spam'
     actions = ['mark_ham', ]
 
 
-class UnsureCommentAdmin(CommentProxyAdmin):
+class UnsureCommentAdmin(CommentAdmin):
     cls = 'unsure'
-    actions = ['mark_ham', 'mark_spam', ]
+    actions = ['add_moderator_reply', 'mark_ham', 'mark_spam', ]
 
 
 class AdminModeratorMixin(object):
